﻿//------------------------------------------------------------------------------------------
//  Instagramへ画像とテキストをアップロードするプログラム
//------------------------------------------------------------------------------------------
//  指定されたINIファイルからアップロード対象として指定されたフォルダよりJPEGファイル/PNG
//  ファイルとテキストファイルを探し、INIファイルで指定されたInstagramのユーザ名/パスワー
//  ドでアップロードを行うプログラムです。
//  PNGファイルの場合、JPEGファイルに変換しアップロードを行います。
//  アーカイブ先が指定されている場合、アップロードしたファイルをアーカイブへ移動し重複投
//  稿を防ぎます。
// 
//  ※注意
//  Instagramへアップロードする画像は事前にInstagramが受け付ける形式への加工が必要です。
//  加工がされていないデータをアップロードすると画像サイズ/アスペクト比の異常などではじか
//  れる場合があるようです。
//  
//  This software is distributed under the license of NYSL.
//  http://www.kmonos.net/nysl/
//  
//  《謝辞》
//  このプログラムで使用しているInstagram APIのラッパーはRamtin Jokarさんが公開して
//  いるInstagramApiSharpを使用して作られております。
//  貴重なプログラムのソースコードを公開されているRamtin Jokarさんに感謝を申し上げま
//  す。
//  
//  入手先:
//  https://github.com/ramtinak/InstagramApiSharp
//  https://www.nuget.org/packages/InstagramApiSharp
//
//  《使い方》
//  InstaUploader.exe <プログラムが使用するINIファイル>
//  
//  INIファイルの指定が省略された場合、InstaUploader.exeが格納されているフォルダの
//  InstaUploader.iniを使用してプログラムの実行を試みます。
//
//  2022/01/24  初版公開
//  2022/01/26  PNGファイルに対応
//              PNGファイルの場合JPEGファイルに変換しアップロード
//  
using InstagramApiSharp.API;
using InstagramApiSharp.API.Builder;
using InstagramApiSharp.Classes;
using InstagramApiSharp.Classes.Models;
using InstagramApiSharp.Logger;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Media.Imaging;

namespace InstaUploader
{
    /// <summary>
    /// 指定されたINIの内容に従い、画像とテキストをInstagramへアップロードするプログラム
    /// </summary>
    public class InstaUploader
    {

        [DllImport("kernel32.dll", CharSet = CharSet.Auto)]
        private static extern int GetPrivateProfileString(string lpAppName, string lpKeyName, string lpDefault, StringBuilder lpReturnedString, uint nSize, string lpFileName);

        /// <summary>
        /// 
        /// </summary>
        /// <param name="args">0: 処理で使用するINIファイルパス。省略された場合EXEの存在するディレクトリのInstaUploader.iniを使用</param>
        /// <returns></returns>
        public static int Main(string[] args)
        {
            return _mainasync(args).Result;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="args"></param>
        /// <returns></returns>
        private static async Task<int> _mainasync(string[] args)
        {
            var exepath = Assembly.GetEntryAssembly().Location;
            var inipath = args.Length > 0 ? args[0] : System.IO.Path.Combine(System.IO.Path.GetDirectoryName(exepath), "InstaUploader.ini");
            return await new InstaUploader().Perform(inipath);
        }

        /// <summary>画像ファイルとして受け入れる拡張子</summary>
        private static readonly string[] IMAGE_EXTENTIONS = new string[] { ".jpg", ".jpeg", ".png" };
        /// <summary>テキストファイルとして受け入れる拡張子</summary>
        private static readonly string[] TEXT_EXTENTIONS = new string[] { ".txt", ".text" };

        /// <summary>InstagramのAPIラッパー</summary>
        private IInstaApi INSTA;


        /// <summary>
        /// 
        /// </summary>
        /// <param name="inipath"></param>
        /// <returns>
        /// 処理結果
        ///     0   正常終了
        ///     1   設定周りでエラー
        ///     2   Instagram周りでエラー
        ///     3   その他のエラー
        /// </returns>
        private async Task<int> Perform(string inipath)
        { 
            // INIの存在確認
            if (!System.IO.File.Exists(inipath))
            {
                Console.WriteLine($"INIファイルが見つかりません(INI-PATH={inipath})");
                return 1;
            }

            // Instagram接続情報のチェック
            string username = GetIni("DEFAULT", "USERNAME", "", inipath).Trim();
            string password = GetIni("DEFAULT", "PASSWORD", "", inipath).Trim();
            if (string.IsNullOrEmpty(username) || string.IsNullOrEmpty(username))
            {
                Console.WriteLine("ユーザ名またはパスワードが設定されていません");
                return 1;
            }

            // Instagramにアップするフォルダのチェック
            string datapath = GetIni("DEFAULT", "DATA_PATH", "", inipath).Trim();
            if (string.IsNullOrEmpty(datapath))
            {
                Console.WriteLine("アップロード対象が未設定です");
                return 1;
            }
            if (!System.IO.Directory.Exists(datapath))
            {
                Console.WriteLine("アップロード対象が存在しません");
                return 3;
            }
            // アップロード後に移動するアーカイブフォルダのチェック
            string archivepath = GetIni("DEFAULT", "ARCHIVE_PATH", "", inipath).Trim();
            if (!string.IsNullOrEmpty(archivepath))
            {
                if (!System.IO.Directory.Exists(archivepath))
                {
                    System.IO.Directory.CreateDirectory(archivepath);
                }
            }

            //--
            // Instagramへ接続
            //--
            if (!await InitInstagram(username, password))
            {
                return 2;
            }
            //--
            // 指定パスの画像とテキストをInstagramへアップロード
            //--
            if (!await Upload(datapath, archivepath))
            {
                Console.WriteLine("アップロードに失敗しました");
                return 2;
            }

            return 0;
        }

        /// <summary>
        /// Instagramの接続環境を初期化
        /// </summary>
        /// <param name="username">Instagramのユーザ名</param>
        /// <param name="password">Instagramの接続パスワード</param>
        /// <returns>
        /// 処理結果: true: 成功 / false: 失敗
        /// </returns>
        private async Task<bool> InitInstagram(string username, string password)
        {
            // InstagramのAPIラッパーを作成
            INSTA = InstaApiBuilder.CreateBuilder()
                                        .SetUser(new UserSessionData() { UserName = username, Password = password })
                                        .UseLogger(new DebugLogger(LogLevel.All))
                                        .Build();

            // Instagramへログインを試みる
            var result = await INSTA.LoginAsync();
            return result.Succeeded;
        }

        /// <summary>
        /// 指定フォルダの内容をInstagramへアップロード
        /// </summary>
        /// <param name="datapath">アップロード対象のデータが格納されているフォルダ</param>
        /// <param name="archivepah">アップロード後のデータを移動するフォルダ(省略可)</param>
        /// <returns></returns>
        private async Task<bool> Upload(string datapath, string archivepah)
        {
            // 指定フォルダの画像ファイルを列挙
            var imageFiles = from x in System.IO.Directory.GetFiles(datapath)
                        where IMAGE_EXTENTIONS.Any(ext => System.IO.Path.GetExtension(x) == ext)
                        orderby x
                        select x;
            if (imageFiles == null || imageFiles.Count() == 0)
            {
                Console.WriteLine("対象画像ファイルなし");
                return true;
            }

            // アップロード用のデータに変換
            var images = new List<InstaImageUpload>();
            var tempFiles = new List<string>();
            foreach (var imageFile in imageFiles)
            {
                // PNGファイルの場合、JPEGに変換
                if (System.IO.Path.GetExtension(imageFile).ToLower() == ".png")
                {
                    var jpegFile = ConvertToJpeg(imageFile);
                    images.Add(new InstaImageUpload { Height = 0, Width = 0, Uri = jpegFile });
                    tempFiles.Add(jpegFile);

                    continue;
                }

                images.Add(new InstaImageUpload { Height = 0, Width = 0, Uri = imageFile});
            }

            // 指定フォルダの中の最初のテキストファイルをキャプションとして準備
            var textFile = (from x in System.IO.Directory.GetFiles(datapath)
                            where TEXT_EXTENTIONS.Any(ext => System.IO.Path.GetExtension(x) == ext)
                            select x).FirstOrDefault();
            string caption = string.IsNullOrEmpty(textFile) ? string.Empty : System.IO.File.ReadAllText(textFile);

            // Instagramへアップロード
            IResult<InstaMedia> result;
            if (images.Count() == 1)
            {
                result = await INSTA.MediaProcessor.UploadPhotoAsync(images.First(), caption);
            }
            else
            {
                result = await INSTA.MediaProcessor.UploadAlbumAsync(images.ToArray(), new InstaVideoUpload[] { }, caption);
            }

            // アップロードしたファイルをアーカイブへ移動
            if (!string.IsNullOrEmpty(archivepah))
            {
                var destpath = System.IO.Path.Combine(archivepah, $"{DateTime.Now:yyyyMMddHHmmss}");
                System.IO.Directory.CreateDirectory(destpath);
                var files = new List<string>();
                files.AddRange(imageFiles);
                files.Add(textFile);
                foreach (var file in files)
                {
                    if (string.IsNullOrEmpty(file)) continue;
                    System.IO.File.Move(file, System.IO.Path.Combine(destpath, System.IO.Path.GetFileName(file)));
                }
            }
            // テンポラリファイルを削除
            tempFiles.ForEach(x =>
            {
                System.IO.File.Delete(x);
            });

            return result.Succeeded ? true : false;
        }

        /// <summary>
        /// PNGファイルをJPEGファイルに変換
        /// </summary>
        /// <param name="pngFilePath">PNGファイルのパス</param>
        /// <returns>変換したJPEGファイルのパス</returns>
        private static string ConvertToJpeg(string pngFilePath)
        {
            // 新しいファイル名はオリジナルのファイル名にGUIDをつけて重複が起こらないようにする
            var jpegFileName = System.IO.Path.GetFileNameWithoutExtension(pngFilePath) + "_" + Guid.NewGuid().ToString() + ".jpg";
            var jpegFilePath = System.IO.Path.Combine(System.IO.Path.GetDirectoryName(pngFilePath), jpegFileName);

            using (var instream = new System.IO.FileStream(pngFilePath, System.IO.FileMode.Open))
            using (var outstream = new System.IO.FileStream(jpegFilePath, System.IO.FileMode.Create))
            {
                var decoder = new PngBitmapDecoder(instream, BitmapCreateOptions.PreservePixelFormat, BitmapCacheOption.Default);
                var encoder = new JpegBitmapEncoder();
                encoder.QualityLevel = 100;

                encoder.Frames.Add(BitmapFrame.Create(decoder.Frames[0]));
                encoder.Save(outstream);
                outstream.Flush();
            }

            return jpegFilePath;
        }

        /// <summary>
        /// INIファイルから文字列の取得
        /// </summary>
        /// <param name="appname"></param>
        /// <param name="keyname"></param>
        /// <param name="defaultvalue"></param>
        /// <param name="filename"></param>
        /// <returns></returns>
        private static string GetIni(string appname, string keyname, string defaultvalue, string filename)
        {
            var sb = new StringBuilder(255);
            GetPrivateProfileString(appname, keyname, defaultvalue, sb, Convert.ToUInt32(sb.Capacity), filename);
            return sb.ToString();
        }
    }
}
